/*============================================
  GView.cpp
  
  1996 Michael Stochosky, see BeVGB.doc for
  details.
  ============================================*/

#include <stdio.h>
#include <string.h>

#ifndef _BITMAP_H
#include <Bitmap.h>
#endif

#ifndef _VIEW_H
#include "View.h"
#endif

#ifndef G_VIEW_H
#include "GView.h"
#endif

#ifndef G_APPLICATION_H
#include "GApplication.h"
#endif

#ifndef G_WINDOW_H
#include "GWindow.h"
#endif

#ifndef G_BITMAP_H
#include "GBitmap.h"
#endif


GView::GView(BRect rect, char *name)
	   	   : BView(rect, name, B_FOLLOW_ALL_SIDES,\
	   	   		B_WILL_DRAW | B_FULL_UPDATE_ON_RESIZE | B_NAVIGABLE )
{
	// create gameboy object and bitmap buffer
	mGameBoy = new GameBoy();
	mBitmap = new BBitmap(BRect(0,0,WIDTH-1,HEIGHT-1), B_COLOR_8_BIT);
	
	// find default file
	mLastRef.database = -1;
	mLastRef.record = -1;
	
	// start out paused
	mPaused = false;
}


GView::~GView()
{
	delete mBitmap;
	delete mGameBoy;
}


// has just been attached to window
void GView::AttachedToWindow()
{
	SetFontName( "Kate" );
	SetFontSize( 9 );
	SetFontRotation( 0.0 );
	SetViewColor(B_TRANSPARENT_32_BIT);
}


// keep emulator from chewing cycles while app is in background
void GView::WindowActivated(bool active)
{
	// resume execution
	if( active )
	{
		// it it wasn't paused when we left, resume it
		if( mPaused == false )
			mGameBoy->Resume();
	}
	
	// halt execution, don't change mPaused flag
	else
		mGameBoy->Suspend();
}


// handle returned drag & drop messages
void GView::MessageReceived(BMessage *msg)
{	
	// get last message
	BMessage *prev = msg->Previous();
	
	// drag message handler
	if( msg->what == B_MESSAGE_NOT_UNDERSTOOD && prev != NIL )
	{		
		// if last message was a 'PPNT' send a 'DATA' instead
		if( prev->what == tIMAGE_VIEWER )		// sent by us and ImageViewer
		{
			record_ref ref;
			
			// create a clipping in the /boot/tmp directory
			// and send event to browser
			BRect source(0, 0, 160-1, 144-1);
			source.OffsetBy((WIDTH-160)/2, 0.0);
			GBitmap *image = new GBitmap(mBitmap, Bounds(), source);
			if( CreateClip(image, &ref) == B_NO_ERROR )
			{
				BMessage *clip = new BMessage(prev);
				
				// remove unwanted fields
				clip->what = B_SIMPLE_DATA;
				clip->RemoveName("bounds");
				clip->RemoveName("color_space");
				clip->RemoveName("color_space");
				clip->RemoveName("bits_length");
				clip->RemoveName("bits");
				
				// add file ref
				clip->AddRef("refs", ref);
				
				// ...right back at ya!
				msg->SendReply(clip);
			}
			
			delete image;
		}
	}
	
	// set gameboy flags
	else if( msg->what == mGB_FLAGS )
		mGameBoy->SetFlags(msg);
	
	// everything else
	else
		inherited::MessageReceived(msg);
}


// handle mouse messages and dragging
void GView::MouseDown(BPoint point)
{
	GBitmap *image;
	BMessage *msg;

	// construct image
	BRect source(0, 0, 160-1, 144-1);
	source.OffsetBy((WIDTH-160)/2, 0.0);
	image = new GBitmap(mBitmap, Bounds(), source);
	
	image->Lock();
	mBitmap->Lock();
	
	// build message
	msg = new BMessage(tIMAGE_VIEWER);		// Painting message sent by ImageViewer
	msg->AddRect("bounds", image->Bounds());
	msg->AddLong("color_space", image->ColorSpace());
	msg->AddLong("bits_length", image->BitsLength());
	msg->AddData("bits", B_RAW_TYPE, image->Bits(), image->BitsLength());
	
	// unlock and send
	mBitmap->Unlock();
	image->Unlock();
	DragMessage(msg, image, point);
}


// drag and drop messages
void GView::MouseMoved(BPoint point, ulong transit, BMessage *msg)
{
	// bail if message is nil
	if( msg == NIL )
		return;
		
	// if it's a DATA message and it has files attached
	if( msg->what == B_SIMPLE_DATA && msg->HasRef("refs") )
	{
		BFile file;
		ulong type, app;
		record_ref ref;
		
		// get the file type
		ref = msg->FindRef("refs");
		file.SetRef(ref);
		file.GetTypeAndApp(&type, &app);
		
		// will accept drags, draw view to reflect validity
		if( type == tCART )
		{
			DrawDragover( transit );
			return;
		}
	}
	else
		inherited::MouseMoved( point, transit, msg);
}


// update view
void GView::Draw(BRect updateRect )
{
	// blit frame buffer to screen
	mBitmap->Lock();
	BRect source(0, 0, 160-1, 144-1);
	source.OffsetBy((WIDTH-160)/2, 0.0);
	DrawBitmap(mBitmap, source, Frame());
	mBitmap->Unlock();
}


// draw border arround view to indicate drag will be accepted
void GView::DrawDragover(ulong transit)
{
	// draw a thin line around edge
	if( transit == B_ENTERED_VIEW )
	{
		rgb_color save = HighColor();
		
		SetHighColor(0, 0, 0);
		StrokeRect(Bounds());
		SetHighColor(save);
	}
	
	if( transit == B_EXITED_VIEW )
		Draw(Bounds());
}


// create a clipping of image in the tmp directory and returns its ref
long GView::CreateClip(BBitmap *image, record_ref *ref)
{
	BVolume boot = boot_volume();
	BDirectory root, tmp;
	BFile clip;
	long err = B_ERROR;
	
	// get tmp directory
	boot.GetRootDirectory(&root);
	root.GetDirectory("tmp", &tmp);
	
	// create clip file
	err = tmp.GetFile("clip", &clip);
	if( err != B_NO_ERROR )
		err = tmp.Create("clip", &clip);
	
	if( err == B_NO_ERROR )
	{	
		// open file and write it out
		err = clip.Open(B_EXCLUSIVE);
		if( err == B_NO_ERROR )
		{
			long width, height, zero;
			color_space color;
			long length;
			long len;
			
			// init
			zero = 0;
			width = image->Bounds().Width();
			height = image->Bounds().Height();
			color = image->ColorSpace();
			length = image->BitsLength();
			image->Lock();
			
			// write bounds, color mode, length of image, and image data
			len = clip.Write(&zero, sizeof(long));
			len = clip.Write(&zero, sizeof(long));
			len = clip.Write(&width, sizeof(long));
			len = clip.Write(&height, sizeof(long));
			len = clip.Write(&color, sizeof(color_space));
			len = clip.Write(&length, sizeof(long));
			len = clip.Write(image->Bits(), image->BitsLength());
			
			// clean up
			image->Unlock();
			err = clip.Close();
			clip.SetTypeAndApp(tIMAGE, tIMAGE_VIEWER);
			
			// asign record ref
			*ref = clip.Record()->Ref();
		}
	}
	return err;
}


// set the bits of the private screen bitmap
void GView::SetBits(char *bits, long length)
{
	if( bits && length > 0 )
	{
		mBitmap->Lock();
		mBitmap->SetBits(bits, length, 0, B_COLOR_8_BIT);
		mBitmap->Unlock();
	}
}


// pause execution of emulator
long GView::Pause(void)
{
	long err;
	
	err = mGameBoy->Suspend();
	if( err == B_NO_ERROR )
		mPaused = true;
	return err;
}


// resume execution of emulator
long GView::Resume(void)
{
	long err;
	
	err = mGameBoy->Resume();	
	if( err == B_NO_ERROR )
		mPaused = false;
	return err;
}


// reset the emulator with a new cartridge
void GView::Run(record_ref ref)
{
	mLastRef = ref;
	mGameBoy->Stop();
	mGameBoy->Go(ref);
	if( Window()->IsActive() )
		mPaused = false;
	else
		mGameBoy->Suspend();
}


// reset the emulator with current cartridge, default if none
void GView::Reset(void)
{
	mGameBoy->Stop();
	mGameBoy->Go(mLastRef);
	mPaused = false;
}


// is execution paused?
bool GView::IsPaused(void)
{
	return mGameBoy->IsSuspended();
}