/*=========================================================================	
  GBGameBoy.cpp
  
  1996 Michael Stochosky, see BeVGB.doc for details. Portions of this
  code are Copyright (C) Marat Fayzullin 1994,1995. Please see GB.c
  for details.
  =========================================================================*/

#ifndef GAMBOY_H
#include "GameBoy.h"
#endif

#ifndef G_APPLICATION_H
#include "GApplication.h"
#endif

extern "C" {
#include "GB.h"

// globals used by emulator
byte gGrayscalePal[12] =	{63,20,10,0, 63,20,10,0, 63,20,10,0};
byte gSpinachPal[12] =		{99,138,176,5, 99,138,176,5, 99,138,176,5};	// 99,137,176,214
byte gDebugPal[12] =		{123,156,42,48, 78,86,55,58, 146,94,32,38};
byte *XPal = gGrayscalePal;
byte *XBuf = NIL;
byte *ZBuf = NIL;
}

extern port_id gOutputPort;
extern port_id gInputPort;


GameBoy::GameBoy(void)
{
	mThread = -1;
	mInteruptSem = -1;
	mExitSem = -1;
	
	InitMachine();
}


GameBoy::~GameBoy( void )
{
	// if thread is running stop
	Stop();
	
	TrashMachine();
}


// run with the specified file
long GameBoy::Go(record_ref ref)
{
	printf("Starting emultor...\n");
	
	long err = B_NO_ERROR;
	
	// create file and start
	BFile *file = new BFile();
	file->SetRef(ref);
	
	if( IsCartridge(ref) )
	{
		mThread = spawn_thread(entry_func, "gb_emulator", \
						B_URGENT_DISPLAY_PRIORITY, file);
		err = resume_thread(mThread);
	}
	
	// not a cartridge
	else
	{
		char name[B_OS_NAME_LENGTH];
		
		file->GetName(name);
		printf("File '%s' is not a GameBoy cartridge.\n", name);
		return B_ERROR;
	}
	
	return err;
}


// stop the gameboy emulator and get it ready for another round
void GameBoy::Stop(void)
{
	printf("Stopping emulator...\n");
	
	thread_info info;
	
	if( get_thread_info( mThread, &info) == B_NO_ERROR )
	{
		// enable the thread in case it was halted
		Resume();
		
		// request thread die
		char *buf = (char*)malloc(2);
		buf[0] = 0xFF;	// no keys down
		buf[1] = true;	// quit requested
		if( write_port_etc(gInputPort, B_RAW_TYPE, buf, 2, B_TIMEOUT, 10000.0) == B_NO_ERROR )
		{
			long err;
			wait_for_thread(mThread, &err);
			
			printf("Emulator exited\n");
		}
		
		// thread not responding, kill it!
		else
		{
			kill_thread(mThread);
			printf("Emulator was killed\n");
		}
		
		
		// release memory occupied by StartGB
		TrashGB();
		
		// clean up
		if( buf ) free(buf);
	}
	else
		printf("Emulator is not running\n");
}


// suspend emulator
long GameBoy::Suspend(void)
{
	return suspend_thread(mThread);
}


// resume emulator thread
long GameBoy::Resume(void)
{
	return resume_thread(mThread);
}


// is the emualtor suspended?
bool GameBoy::IsSuspended(void)
{
	thread_info info;
	
	if( get_thread_info(mThread, &info) == B_NO_ERROR )
	{
		if( info.state == B_THREAD_SUSPENDED )
			return true;
		else
			return false;
	}
	else
		return true;
}


// is file an executable cartridge?
bool GameBoy::IsCartridge(record_ref ref)
{
	BFile file;
	
	// open file
	file.SetRef(ref);
	if( file.Open(B_READ_ONLY) == B_NO_ERROR )
	{
		unsigned char RAM[0x4000];
		unsigned checksum;
		int i;
		
		// count down
		if( file.Read(RAM, 0x4000) == 0x4000 )
		{	
			checksum = ((unsigned)RAM[0x014E] << 8) + RAM[0x014F];
			checksum += RAM[0x014E] + RAM[0x014F];
			
			do{
				for(i=0; i<0x4000; i++)
					checksum -= RAM[i];
			}while( file.Read(RAM, 0x4000) == 0x4000 );
		}
		else
			checksum = -1;
		
		// clean up
		file.Close();
		return !(checksum & 0xFFFF);
	}
	else
		return false;
}


void GameBoy::SetFlags(BMessage *msg)
{
	// set the UPeriod
	if( msg->HasLong("UPeriod") )
		UPeriod = msg->FindLong("UPeriod");
	
	// set the palette
	if( msg->HasLong("palette") )
	{
		long index = msg->FindLong("palette");
		switch( index )
		{
			case 0: XPal = gGrayscalePal; break;
			case 1: XPal = gSpinachPal; break;
			case 2: XPal = gDebugPal; break;
			default: XPal = gGrayscalePal;
		}
	}
}


// entry point for emulator thread
long GameBoy::entry_func(void *arg)
{
	BFile *file = (BFile*)arg;
	char path[512];
	char name[B_OS_NAME_LENGTH];
	
	// get name
	file->GetName(name);
	
	// set signature of file to that of a cartridge
	file->SetTypeAndApp(tCART, tSIGNATURE);
	
	// get path of cartridge
	if( file->GetPath(path, 512) == B_NO_ERROR )
	{
		if( StartGB(path) )
		{
			char *savename;
			
			// get save files name
			savename = (char*)malloc(strlen(name)+10);
			if( savename != NIL )
			{				
				char *str;
				strcpy(savename, name);
				str = strrchr(savename, '.');
				if( str != NIL )
					strcpy(str, ".sav");
				else
					strcat(savename, ".sav");
				
				// set save files signature
				BDirectory dir;
				BFile savefile;
				file->GetParent(&dir);
				if( dir.GetFile(savename, &savefile) == B_NO_ERROR )
					savefile.SetTypeAndApp(tSAVE, tSIGNATURE);
				
				// clean up
				free(savename);
			}
		}// exited okay
	}
	else
		return B_ERROR;
		
	return B_NO_ERROR;
}

// Declared in GB.h -------------------------------------------------------
#pragma mark -

int InitMachine()
{	
	// alloc buffers
	if( XBuf == NIL ) XBuf = (byte*)malloc(WIDTH*HEIGHT);
	if( ZBuf == NIL ) ZBuf = (byte*)malloc(2*WIDTH*HEIGHT/8);
	
	return 1;
}
 
void TrashMachine(void)
{
	// delete buffers and set globals to nil
	if( XBuf ){ free(XBuf); XBuf = NIL; }
	if( ZBuf ){ free(ZBuf); XBuf = NIL; }
}

void RefreshScreen(void)
{
	// send new image
	write_port_etc(gOutputPort, B_RAW_TYPE, XBuf, WIDTH*HEIGHT, B_TIMEOUT, 30000.0);
}

#ifdef USE_XPAL
#define XPAL(N) XPal[N]
#else
#define XPAL(N) N
#endif


// Refresh a line, from 0 to 143
void RefreshLine(byte Y)
{
  register byte Offset,*P,*T,*Z,Z0,D0,X1,X2,Shift;
  PIXEL *R;
  word D1;

#ifdef USE_XPAL
#define PAL(N)  Pal[N]
#define WPAL(N) WPal[N]
  PIXEL Pal[4],WPal[4];
  Pal[0]=XPal[BPal[0]];Pal[1]=XPal[BPal[1]];
  Pal[2]=XPal[BPal[2]];Pal[3]=XPal[BPal[3]];
  WPal[0]=(XPal+8)[BPal[0]];WPal[1]=(XPal+8)[BPal[1]];
  WPal[2]=(XPal+8)[BPal[2]];WPal[3]=(XPal+8)[BPal[3]];
#else
#define PAL(N)  BPal[N]
#define WPAL(N) BPal[N]+8
#endif

  D1=WIDTH*Y+(WIDTH-160)/2;
  R=XBuf+D1;Z=ZBuf+(D1>>3);

  if((LCDCONT&0x81)!=0x81)
  {
    /* Clearing screen buffer */
    memset(R,XPAL(0),160);
    /* Clearing Z-buffer */
    memset(Z,0x00,160/8);
    memset(Z+WIDTH*HEIGHT/8,0x00,160/8);
  }
  else
  {
    Offset=Y+SCROLLY;
    T=BgdTab+((short)(Offset&0xF8)<<2);
    Offset=(Offset&0x07)<<1;

    Shift=SCROLLX&0x07;
    R-=Shift;Z--;
    X1=SCROLLX>>3;

    /* Determine how much background is shown */
    X2=(WNDPOSX>7)? WNDPOSX-7:0;
    if((~LCDCONT&0x20)||(WNDPOSY>Y)||(X2>159)) X2=160;
    X2=X1+((X2+7)>>3);

    Z[0]=0x00;

    for(;X1<=X2;X1++)
    {
      D0=*(T+(X1&0x1F));
      if(~LCDCONT&0x10) D0+=0x80;
      P=ChrGen+(D0<<4)+Offset;
      D0=*P;

      /* Modifying Z-buffer */
      D1=(short)(D0|*(P+1))<<Shift;
      Z[0]|=D1>>8;
      Z[1]=D1&0xFF;
      Z[WIDTH*HEIGHT/8]=0x00;

      /* Modifying screen buffer */
      D1=(short)*(P+1)<<1;
      D1=(D1&0xAAA)|((D1&0x555)<<7)|(D0&0x55)|((short)(D0&0xAA)<<7);
      R[0]=PAL((D1&0xC000)>>14);
      R[1]=PAL((D1&0x00C0)>>6 );
      R[2]=PAL((D1&0x3000)>>12);
      R[3]=PAL((D1&0x0030)>>4 );
      R[4]=PAL((D1&0x0C00)>>10);
      R[5]=PAL((D1&0x000C)>>2 );
      R[6]=PAL((D1&0x0300)>>8 );
      R[7]=PAL( D1&0x003      );

      /* Moving pointers */
      R+=8;Z++;
    }
  }

  X1=(WNDPOSX>7)? WNDPOSX-7:0;
  if(((LCDCONT&0xA0)==0xA0)&&(WNDPOSY<=Y)&&(X1<160))
  {
    D1=WIDTH*Y+(WIDTH-160)/2+X1;
    R=XBuf+D1;Z=ZBuf+(D1>>3);

    Offset=Y-WNDPOSY;
    T=WndTab+((short)(Offset&0xF8)<<2);
    Offset=(Offset&0x07)<<1;

    D0=0xFF>>(X1&0x07);
    if(LCDCONT&0x02) Z[WIDTH*HEIGHT/8]=Z0=0x00;
    else { Z[WIDTH*HEIGHT/8]=D0;Z0=0xFF; }
    Z[0]|=D0;Z++;

    for(X1>>=3;X1<(160>>3);X1++)
    {
      D0=*(T++)+0x80;
      P=RAM+0x8800+((short)D0<<4)+Offset;

      /* Modifying Z-buffer */
      Z[WIDTH*HEIGHT/8]=Z0;Z[0]|=0xFF;

      /* Modifying screen buffer */
      D0=*P;D1=(short)*(P+1)<<1;
      D1=(D1&0xAAA)|((D1&0x555)<<7)|(D0&0x55)|((short)(D0&0xAA)<<7);
      R[0]=WPAL((D1&0xC000)>>14);
      R[1]=WPAL((D1&0x00C0)>>6 );
      R[2]=WPAL((D1&0x3000)>>12);
      R[3]=WPAL((D1&0x0030)>>4 );
      R[4]=WPAL((D1&0x0C00)>>10);
      R[5]=WPAL((D1&0x000C)>>2 );
      R[6]=WPAL((D1&0x0300)>>8 );
      R[7]=WPAL( D1&0x0003     );

      /* Moving pointers */
      R+=8;Z++;
    }
  }
}

void RefreshSprites(void)
{
  /* Conversion matrix where result has bits */
  /* flipped over and inverted.              */
  static byte Conv[256] =
  {
    0xFF,0x7F,0xBF,0x3F,0xDF,0x5F,0x9F,0x1F,0xEF,0x6F,0xAF,0x2F,
    0xCF,0x4F,0x8F,0x0F,0xF7,0x77,0xB7,0x37,0xD7,0x57,0x97,0x17,
    0xE7,0x67,0xA7,0x27,0xC7,0x47,0x87,0x07,0xFB,0x7B,0xBB,0x3B,
    0xDB,0x5B,0x9B,0x1B,0xEB,0x6B,0xAB,0x2B,0xCB,0x4B,0x8B,0x0B,
    0xF3,0x73,0xB3,0x33,0xD3,0x53,0x93,0x13,0xE3,0x63,0xA3,0x23,
    0xC3,0x43,0x83,0x03,0xFD,0x7D,0xBD,0x3D,0xDD,0x5D,0x9D,0x1D,
    0xED,0x6D,0xAD,0x2D,0xCD,0x4D,0x8D,0x0D,0xF5,0x75,0xB5,0x35,
    0xD5,0x55,0x95,0x15,0xE5,0x65,0xA5,0x25,0xC5,0x45,0x85,0x05,
    0xF9,0x79,0xB9,0x39,0xD9,0x59,0x99,0x19,0xE9,0x69,0xA9,0x29,
    0xC9,0x49,0x89,0x09,0xF1,0x71,0xB1,0x31,0xD1,0x51,0x91,0x11,
    0xE1,0x61,0xA1,0x21,0xC1,0x41,0x81,0x01,0xFE,0x7E,0xBE,0x3E,
    0xDE,0x5E,0x9E,0x1E,0xEE,0x6E,0xAE,0x2E,0xCE,0x4E,0x8E,0x0E,
    0xF6,0x76,0xB6,0x36,0xD6,0x56,0x96,0x16,0xE6,0x66,0xA6,0x26,
    0xC6,0x46,0x86,0x06,0xFA,0x7A,0xBA,0x3A,0xDA,0x5A,0x9A,0x1A,
    0xEA,0x6A,0xAA,0x2A,0xCA,0x4A,0x8A,0x0A,0xF2,0x72,0xB2,0x32,
    0xD2,0x52,0x92,0x12,0xE2,0x62,0xA2,0x22,0xC2,0x42,0x82,0x02,
    0xFC,0x7C,0xBC,0x3C,0xDC,0x5C,0x9C,0x1C,0xEC,0x6C,0xAC,0x2C,
    0xCC,0x4C,0x8C,0x0C,0xF4,0x74,0xB4,0x34,0xD4,0x54,0x94,0x14,
    0xE4,0x64,0xA4,0x24,0xC4,0x44,0x84,0x04,0xF8,0x78,0xB8,0x38,
    0xD8,0x58,0x98,0x18,0xE8,0x68,0xA8,0x28,0xC8,0x48,0x88,0x08,
    0xF0,0x70,0xB0,0x30,0xD0,0x50,0x90,0x10,0xE0,0x60,0xA0,0x20,
    0xC0,0x40,0x80,0x00
  };

  byte *T,*S,*R,*Z,I,J,K,N,D0;
  word D1;
  short Shift,DY;
  PIXEL *P;

#undef PAL
#ifdef USE_XPAL
#define PAL(N) Pal[N]
  PIXEL Pal[4];
#else
#define PAL(N) R[N]+4
#endif

  N=LCDCONT&0x04? 16:8;

  for(S=RAM+0xFE9C,J=40;J;S-=4,J--)
    if(S[0]&&(S[0]<160)&&S[1]&&(S[1]<168))
    {
      if(S[0]<16) { D0=0;K=16-S[0];I=(K<N)? N-K:0; }
      else        { D0=S[0]-16;K=0;I=(160-S[0]<N)? 160-S[0]:N; }
      Shift=S[1]-(S[3]&0x20? 1:8);
      D1=D0*WIDTH+(WIDTH-160)/2+Shift;
      P=XBuf+D1;Z=ZBuf+(D1>>3);
      Shift&=0x07;
      DY=2;
      if(S[3]&0x40) { DY=-DY;K=N-K-1; }
      R=S[3]&0x10? SPal1:SPal0;
#ifdef USE_XPAL
      Pal[0]=(XPal+4)[R[0]];Pal[1]=(XPal+4)[R[1]];
      Pal[2]=(XPal+4)[R[2]];Pal[3]=(XPal+4)[R[3]];
#endif
      T=RAM+0x8000+((long)((N>8)? S[2]&0xFE:S[2])<<4)+(K<<1);

      /* If on top of background, clip only against the window */
      if(~S[3]&0x80) Z+=WIDTH*HEIGHT/8;

      if(S[3]&0x20)
      {
        Shift=7-Shift;
        for(;I;I--,P+=WIDTH,Z+=WIDTH/8,T+=DY)
        {
          D0=*T;D1=*(T+1);
          K=(((short)Z[-1]<<8)|Z[0])>>Shift;
          K=(D0|D1)&Conv[K];
          if(K)
          {
            D1<<=1;
            D1=(D1&0xAAA)|((D1&0x555)<<7)|(D0&0x55)|((short)(D0&0xAA)<<7);
            if(K&0x80) P[0]=PAL((D1&0xC000)>>14);
            if(K&0x40) P[-1]=PAL((D1&0x00C0)>>6 );
            if(K&0x20) P[-2]=PAL((D1&0x3000)>>12);
            if(K&0x10) P[-3]=PAL((D1&0x0030)>>4 );
            if(K&0x08) P[-4]=PAL((D1&0x0C00)>>10);
            if(K&0x04) P[-5]=PAL((D1&0x000C)>>2 );
            if(K&0x02) P[-6]=PAL((D1&0x0300)>>8 );
            if(K&0x01) P[-7]=PAL( D1&0x0003     );
          }
        }
      }
      else
      {
        Shift=8-Shift;
        for(;I;I--,P+=WIDTH,Z+=WIDTH/8,T+=DY)
        {
          D0=*T;D1=*(T+1);
          K=(((short)Z[0]<<8)|Z[1])>>Shift;
          K=(D0|D1)&~K;
          if(K)
          {
            D1<<=1;
            D1=(D1&0xAAA)|((D1&0x555)<<7)|(D0&0x55)|((short)(D0&0xAA)<<7);
            if(K&0x80) P[0]=PAL((D1&0xC000)>>14);
            if(K&0x40) P[1]=PAL((D1&0x00C0)>>6 );
            if(K&0x20) P[2]=PAL((D1&0x3000)>>12);
            if(K&0x10) P[3]=PAL((D1&0x0030)>>4 );
            if(K&0x08) P[4]=PAL((D1&0x0C00)>>10);
            if(K&0x04) P[5]=PAL((D1&0x000C)>>2 );
            if(K&0x02) P[6]=PAL((D1&0x0300)>>8 );
            if(K&0x01) P[7]=PAL( D1&0x0003     );
          }
        }
      }
    }
}


// pole the state of the keyboard for input
extern byte JoyState;
byte Joystick(void)
{
	long type, length;
	char buf[2];
	static byte joy = JoyState;
	
	// get joystick state
	length = read_port_etc(gInputPort, &type, buf, 2, B_TIMEOUT, 0.0);
	if( (length == 2) && (type == B_RAW_TYPE) )
	{
		// if we've been requested to quit tell emulator to exit
		if( buf[1] == true )
			CPURunning = 0;		// kill Z80
		
		joy = buf[0];
	}
	
	return joy;
}


// push sound into sound channel
void Sound(byte R,byte V)
{
	return;
}


// send byte to serial line
byte SIOSend(byte V)
{
	return 0;
}


// get byte from serial line
byte SIOReceive(byte *V)
{
	return 0;
}